# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.common as pdc
from pydiverse.common import (
    Bool,
    Date,
    Datetime,
    Dtype,
    Float,
    Float32,
    Float64,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    String,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)
from pydiverse.common.testing import ALL_TYPES

try:
    import polars as pl
except ImportError:
    pl = None


@pytest.mark.skipif(pl is None, reason="requires polars")
def test_dtype_from_polars():
    def assert_conversion(type_, expected):
        assert Dtype.from_polars(type_) == expected

    assert_conversion(pl.Int64, Int64())
    assert_conversion(pl.Int32, Int32())
    assert_conversion(pl.Int16, Int16())
    assert_conversion(pl.Int8, Int8())

    assert_conversion(pl.UInt64, UInt64())
    assert_conversion(pl.UInt32, UInt32())
    assert_conversion(pl.UInt16, UInt16())
    assert_conversion(pl.UInt8, UInt8())

    assert_conversion(pl.Float64, Float64())
    assert_conversion(pl.Float32, Float32())

    assert_conversion(pl.Utf8, String())
    assert_conversion(pl.Boolean, Bool())

    assert_conversion(pl.Date, Date())
    assert_conversion(pl.Time, Time())
    assert_conversion(pl.Datetime, Datetime())
    assert_conversion(pl.Datetime("ms"), Datetime())
    assert_conversion(pl.Datetime("us"), Datetime())
    assert_conversion(pl.Datetime("ns"), Datetime())


@pytest.mark.skipif(pl is None, reason="requires polars")
def test_dtype_to_polars():
    def assert_conversion(type_: Dtype, expected):
        assert type_.to_polars() == expected

    assert_conversion(Int64(), pl.Int64)
    assert_conversion(Int32(), pl.Int32)
    assert_conversion(Int16(), pl.Int16)
    assert_conversion(Int8(), pl.Int8)

    assert_conversion(UInt64(), pl.UInt64)
    assert_conversion(UInt32(), pl.UInt32)
    assert_conversion(UInt16(), pl.UInt16)
    assert_conversion(UInt8(), pl.UInt8)

    assert_conversion(Float64(), pl.Float64)
    assert_conversion(Float32(), pl.Float32)

    assert_conversion(String(), pl.Utf8)
    assert_conversion(Bool(), pl.Boolean)

    assert_conversion(Date(), pl.Date)
    assert_conversion(Time(), pl.Time)
    assert_conversion(Datetime(), pl.Datetime("us"))


@pytest.mark.skipif(pl is None, reason="requires polars")
@pytest.mark.parametrize(
    "type_",
    ALL_TYPES,
)
def test_all_types(type_):
    if type_ is pdc.List:
        type_obj = type_(pdc.Int64())
    else:
        type_obj = type_()

    acceptance_map = {
        Float: Float64(),
        Int: Int64(),
    }

    dst_type = type_obj.to_polars()
    back_type = Dtype.from_polars(dst_type)
    assert back_type == acceptance_map.get(type_, type_obj)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import logging
import os

import pytest

from pydiverse.common.util.structlog import setup_logging

log_level = logging.INFO if not os.environ.get("DEBUG", "") else logging.DEBUG
setup_logging(log_level=log_level)

try:
    import structlog

    # Pytest Configuration
    @pytest.fixture(autouse=True, scope="function")
    def structlog_test_info(request):
        """Add testcase information to structlog context"""
        if not os.environ.get("DEBUG", ""):
            yield
            return

        with structlog.contextvars.bound_contextvars(testcase=request.node.name):
            yield

    logger = structlog.get_logger()
except ImportError:
    logger = logging.getLogger(__name__)

logger.info("Logging initialized")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import pytest

import pydiverse.common as pdc
from pydiverse.common import (
    Bool,
    Date,
    Datetime,
    Decimal,
    Dtype,
    Enum,
    Float,
    Float32,
    Float64,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    String,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)
from pydiverse.common.testing import ALL_TYPES

try:
    import sqlalchemy as sa
except ImportError:
    sa = None


@pytest.mark.skipif(sa is None, reason="requires sqlalchemy")
def test_dtype_from_sqlalchemy():
    def assert_conversion(type_, expected):
        assert Dtype.from_sql(type_) == expected

    assert_conversion(sa.BigInteger(), Int64())
    assert_conversion(sa.Integer(), Int32())
    assert_conversion(sa.SmallInteger(), Int16())

    assert_conversion(sa.Numeric(), Float64())
    assert_conversion(sa.Numeric(13, 2), Float64())
    assert_conversion(sa.Numeric(1, 0), Float64())
    assert_conversion(sa.DECIMAL(13, 2), Float64())
    assert_conversion(sa.DECIMAL(1, 0), Float64())
    assert_conversion(sa.Float(), Float64())
    assert_conversion(sa.Float(24), Float32())
    assert_conversion(sa.Float(53), Float64())

    assert_conversion(sa.String(), String())
    assert_conversion(sa.Boolean(), Bool())

    assert_conversion(sa.Date(), Date())
    assert_conversion(sa.Time(), Time())
    assert_conversion(sa.DateTime(), Datetime())


@pytest.mark.skipif(sa is None, reason="requires sqlalchemy")
def test_dtype_to_sqlalchemy():
    def assert_conversion(type_: Dtype, expected):
        assert isinstance(type_.to_sql(), expected)

    assert_conversion(Int64(), sa.BigInteger)
    assert_conversion(Int32(), sa.Integer)
    assert_conversion(Int16(), sa.SmallInteger)
    assert_conversion(Int8(), sa.SmallInteger)

    assert_conversion(UInt64(), sa.BigInteger)
    assert_conversion(UInt32(), sa.BigInteger)
    assert_conversion(UInt16(), sa.Integer)
    assert_conversion(UInt8(), sa.SmallInteger)

    assert_conversion(Float64(), sa.Float)
    assert_conversion(Float32(), sa.Float)

    assert_conversion(String(), sa.String)
    assert_conversion(Bool(), sa.Boolean)

    assert_conversion(Date(), sa.Date)
    assert_conversion(Time(), sa.Time)
    assert_conversion(Datetime(), sa.DateTime)


@pytest.mark.skipif(sa is None, reason="requires sqlalchemy")
@pytest.mark.parametrize(
    "type_",
    ALL_TYPES,
)
def test_all_types(type_):
    if type_ is pdc.List:
        type_obj = type_(pdc.Int64())
    elif type_ is pdc.Enum:
        type_obj = type_("a", "b", "c")
    else:
        type_obj = type_()
    dst_type = type_obj.to_sql()
    back_type = Dtype.from_sql(dst_type)
    acceptance_map = {
        # SQL is a bit less strict about integer precisions
        Int8: Int16(),
        UInt8: Int16(),
        UInt16: Int32(),
        UInt32: Int64(),
        UInt64: Int64(),
        # we intentionally fetch Decimal as Float since Decimal is more a relational
        # database thing
        Decimal: Float64(),
        Float: Float64(),
        Int: Int64(),
        # there is no Enum
        Enum: String(),
    }
    assert back_type == acceptance_map.get(type_, type_obj)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import enum


class PandasBackend(str, enum.Enum):
    NUMPY = "numpy"
    ARROW = "arrow"


class Dtype:
    """Base class for all data types."""

    def __eq__(self, rhs):
        """Return ``True`` if this dtype is equal to `rhs`."""
        return isinstance(rhs, Dtype) and type(self) is type(rhs)

    def __hash__(self):
        """Return a hash for this dtype."""
        return hash(type(self))

    def __repr__(self):
        """Return a string representation of this dtype."""
        return self.__class__.__name__

    @classmethod
    def is_int(cls):
        """Return ``True`` if this dtype is an integer type."""
        return False

    @classmethod
    def is_float(cls):
        """Return ``True`` if this dtype is a float type."""
        return False

    @classmethod
    def is_subtype(cls, rhs):
        """Return ``True`` if this dtype is a subtype of `rhs`.

        For example, ``Int8.is_subtype(Int())`` is ``True``.
        """
        rhs_cls = type(rhs)
        return (
            (cls is rhs_cls)
            or (rhs_cls is Int and cls.is_int())
            or (rhs_cls is Float and cls.is_float())
        )

    @staticmethod
    def from_sql(sql_type) -> "Dtype":
        """Convert a SQL type to a Dtype."""
        import sqlalchemy as sqa

        if isinstance(sql_type, sqa.SmallInteger):
            return Int16()
        if isinstance(sql_type, sqa.BigInteger):
            return Int64()
        if isinstance(sql_type, sqa.Integer):
            return Int32()
        if isinstance(sql_type, sqa.Float):
            precision = sql_type.precision or 53
            if precision <= 24:
                return Float32()
            return Float64()
        if isinstance(sql_type, sqa.Numeric | sqa.DECIMAL):
            # Just to be safe, we always use FLOAT64 for fixpoint numbers.
            # Databases are obsessed about fixpoint. However, in dataframes, it
            # is more common to just work with double precision floating point.
            # We see Decimal as subtype of Float. Pydiverse.transform will convert
            # Decimal to Float64 whenever it cannot guarantee semantic correctness
            # otherwise.
            return Float64()
        if isinstance(sql_type, sqa.String):
            return String()
        if isinstance(sql_type, sqa.Boolean):
            return Bool()
        if isinstance(sql_type, sqa.Date):
            return Date()
        if isinstance(sql_type, sqa.Time):
            return Time()
        if isinstance(sql_type, sqa.DateTime):
            return Datetime()
        if isinstance(sql_type, sqa.Interval):
            return Duration()
        if isinstance(sql_type, sqa.ARRAY):
            return List(Dtype.from_sql(sql_type.item_type))
        if isinstance(sql_type, sqa.types.NullType):
            return NullType()

        raise TypeError

    @staticmethod
    def from_pandas(pandas_type) -> "Dtype":
        """Convert a pandas type to a Dtype."""
        import numpy as np
        import pandas as pd

        if isinstance(pandas_type, pd.ArrowDtype):
            return Dtype.from_arrow(pandas_type.pyarrow_dtype)

        def is_np_dtype(type_, np_dtype):
            return pd.core.dtypes.common._is_dtype_type(
                type_, pd.core.dtypes.common.classes(np_dtype)
            )

        workaround = (
            pandas_type is not np.floating
        )  # see https://github.com/pandas-dev/pandas/issues/62018
        if workaround and pd.api.types.is_signed_integer_dtype(pandas_type):
            if is_np_dtype(pandas_type, np.int64):
                return Int64()
            elif is_np_dtype(pandas_type, np.int32):
                return Int32()
            elif is_np_dtype(pandas_type, np.int16):
                return Int16()
            elif is_np_dtype(pandas_type, np.int8):
                return Int8()
            raise TypeError
        if workaround and pd.api.types.is_unsigned_integer_dtype(pandas_type):
            if is_np_dtype(pandas_type, np.uint64):
                return UInt64()
            elif is_np_dtype(pandas_type, np.uint32):
                return UInt32()
            elif is_np_dtype(pandas_type, np.uint16):
                return UInt16()
            elif is_np_dtype(pandas_type, np.uint8):
                return UInt8()
            raise TypeError
        if not workaround or pd.api.types.is_float_dtype(pandas_type):
            if not workaround or is_np_dtype(pandas_type, np.float64):
                return Float64()
            elif is_np_dtype(pandas_type, np.float32):
                return Float32()
            raise TypeError
        if pd.api.types.is_string_dtype(pandas_type):
            # We reserve the use of the object column for string.
            return String()
        if pd.api.types.is_bool_dtype(pandas_type):
            return Bool()
        if pd.api.types.is_datetime64_any_dtype(pandas_type):
            return Datetime()
        if pd.api.types.is_timedelta64_dtype(pandas_type):
            return Duration()
        # we don't know any decimal/time/null dtypes in pandas if column is not
        # arrow backed

        if pandas_type.name == "category":
            return Enum(*pandas_type.categories.to_list())

        raise TypeError

    @staticmethod
    def from_arrow(arrow_type) -> "Dtype":
        """Convert a PyArrow type to a Dtype."""
        import pyarrow as pa

        if pa.types.is_signed_integer(arrow_type):
            if pa.types.is_int64(arrow_type):
                return Int64()
            if pa.types.is_int32(arrow_type):
                return Int32()
            if pa.types.is_int16(arrow_type):
                return Int16()
            if pa.types.is_int8(arrow_type):
                return Int8()
            raise TypeError
        if pa.types.is_unsigned_integer(arrow_type):
            if pa.types.is_uint64(arrow_type):
                return UInt64()
            if pa.types.is_uint32(arrow_type):
                return UInt32()
            if pa.types.is_uint16(arrow_type):
                return UInt16()
            if pa.types.is_uint8(arrow_type):
                return UInt8()
            raise TypeError
        if pa.types.is_floating(arrow_type):
            if pa.types.is_float64(arrow_type):
                return Float64()
            if pa.types.is_float32(arrow_type):
                return Float32()
            if pa.types.is_float16(arrow_type):
                return Float32()
            raise TypeError
        if pa.types.is_decimal(arrow_type):
            # We don't recommend using Decimal in dataframes, but we support it.
            return Decimal()
        if pa.types.is_string(arrow_type):
            return String()
        if pa.types.is_boolean(arrow_type):
            return Bool()
        if pa.types.is_timestamp(arrow_type):
            return Datetime()
        if pa.types.is_date(arrow_type):
            return Date()
        if pa.types.is_time(arrow_type):
            return Time()
        if pa.types.is_duration(arrow_type):
            return Duration()
        if pa.types.is_null(arrow_type):
            return NullType()
        if pa.types.is_list(arrow_type):
            return List(Dtype.from_arrow(arrow_type.value_type))
        if pa.types.is_dictionary(arrow_type):
            raise RuntimeError(
                "Most likely this is an Enum type. But metadata about categories is "
                "only in the pyarrow field and not in the pyarrow dtype"
            )
        raise TypeError

    @staticmethod
    def from_polars(polars_type) -> "Dtype":
        """Convert a Polars type to a Dtype."""
        import polars as pl

        if isinstance(polars_type, pl.List):
            return List(Dtype.from_polars(polars_type.inner))
        if isinstance(polars_type, pl.Enum):
            return Enum(*polars_type.categories)

        return {
            pl.Int64: Int64(),
            pl.Int32: Int32(),
            pl.Int16: Int16(),
            pl.Int8: Int8(),
            pl.UInt64: UInt64(),
            pl.UInt32: UInt32(),
            pl.UInt16: UInt16(),
            pl.UInt8: UInt8(),
            pl.Float64: Float64(),
            pl.Float32: Float32(),
            pl.Decimal: Decimal(),
            pl.Utf8: String(),
            pl.Boolean: Bool(),
            pl.Datetime: Datetime(),
            pl.Time: Time(),
            pl.Date: Date(),
            pl.Null: NullType(),
            pl.Duration: Duration(),
        }[polars_type.base_type()]

    def to_sql(self):
        """Convert this Dtype to a SQL type."""
        import sqlalchemy as sqa

        return {
            Int(): sqa.BigInteger(),  # we default to 64 bit
            Int8(): sqa.SmallInteger(),
            Int16(): sqa.SmallInteger(),
            Int32(): sqa.Integer(),
            Int64(): sqa.BigInteger(),
            UInt8(): sqa.SmallInteger(),
            UInt16(): sqa.Integer(),
            UInt32(): sqa.BigInteger(),
            UInt64(): sqa.BigInteger(),
            Float(): sqa.Float(53),  # we default to 64 bit
            Float32(): sqa.Float(24),
            Float64(): sqa.Float(53),
            Decimal(): sqa.DECIMAL(),
            String(): sqa.String(),
            Bool(): sqa.Boolean(),
            Date(): sqa.Date(),
            Time(): sqa.Time(),
            Datetime(): sqa.DateTime(),
            Duration(): sqa.Interval(),
            NullType(): sqa.types.NullType(),
        }[self]

    def to_pandas(self, backend: PandasBackend = PandasBackend.ARROW):
        """Convert this Dtype to a pandas type."""
        import pandas as pd

        if backend == PandasBackend.NUMPY:
            return self.to_pandas_nullable(backend)
        if backend == PandasBackend.ARROW:
            if self == String() or isinstance(self, Enum):
                return pd.StringDtype(storage="pyarrow")
            return pd.ArrowDtype(self.to_arrow())

    def to_pandas_nullable(self, backend: PandasBackend = PandasBackend.ARROW):
        """Convert this Dtype to a pandas nullable type.

        Nullable can be either pandas extension types like StringDtype or ArrowDtype.

        Parameters
        ----------
        backend : PandasBackend, optional
            The pandas backend to use. Defaults to ``PandasBackend.ARROW``.
            If ``PandasBackend.NUMPY`` is selected, this method will attempt
            to return a NumPy-backed nullable pandas dtype. Note that
            Time, NullType, and List will raise a TypeError for the
            NUMPY backend as pandas doesn't have corresponding native
            nullable dtypes for these.
        """
        import pandas as pd

        if backend == PandasBackend.ARROW:
            return pd.ArrowDtype(self.to_arrow())

        # we don't want to produce object columns
        if isinstance(self, Time):
            raise TypeError("pandas doesn't have a native time dtype")
        if isinstance(self, NullType):
            raise TypeError("pandas doesn't have a native null dtype")
        if isinstance(self, List):
            raise TypeError("pandas doesn't have a native list dtype")

        if isinstance(self, Enum):
            return pd.CategoricalDtype(self.categories)

        return {
            Int(): pd.Int64Dtype(),  # we default to 64 bit
            Int8(): pd.Int8Dtype(),
            Int16(): pd.Int16Dtype(),
            Int32(): pd.Int32Dtype(),
            Int64(): pd.Int64Dtype(),
            UInt8(): pd.UInt8Dtype(),
            UInt16(): pd.UInt16Dtype(),
            UInt32(): pd.UInt32Dtype(),
            UInt64(): pd.UInt64Dtype(),
            Float(): pd.Float64Dtype(),  # we default to 64 bit
            Float32(): pd.Float32Dtype(),
            Float64(): pd.Float64Dtype(),
            Decimal(): pd.Float64Dtype(),  # NumericDtype is
            String(): pd.StringDtype(),
            Bool(): pd.BooleanDtype(),
            Date(): "datetime64[s]",
            Datetime(): "datetime64[us]",
            Time(): "timedelta64[us]",
            Duration(): "timedelta64[us]",
        }[self]

    def to_arrow(self):
        """Convert this Dtype to a PyArrow type."""
        import pyarrow as pa

        if isinstance(self, Enum):
            return pa.string()

        return {
            Int(): pa.int64(),  # we default to 64 bit
            Int8(): pa.int8(),
            Int16(): pa.int16(),
            Int32(): pa.int32(),
            Int64(): pa.int64(),
            UInt8(): pa.uint8(),
            UInt16(): pa.uint16(),
            UInt32(): pa.uint32(),
            UInt64(): pa.uint64(),
            Float(): pa.float64(),  # we default to 64 bit
            Float32(): pa.float32(),
            Float64(): pa.float64(),
            Decimal(): pa.decimal128(35, 10),  # Arbitrary precision
            String(): pa.string(),
            Bool(): pa.bool_(),
            Date(): pa.date32(),
            Time(): pa.time64("us"),
            Datetime(): pa.timestamp("us"),
            Duration(): pa.duration("us"),
            NullType(): pa.null(),
        }[self]

    def to_arrow_field(self, name: str, nullable: bool = True):
        """Convert this Dtype to a PyArrow Field."""
        import pyarrow as pa

        return pa.field(name, self.to_arrow(), nullable=nullable)

    def to_polars(self: "Dtype"):
        """Convert this Dtype to a Polars type."""
        import polars as pl

        return {
            Int(): pl.Int64,  # we default to 64 bit
            Int64(): pl.Int64,
            Int32(): pl.Int32,
            Int16(): pl.Int16,
            Int8(): pl.Int8,
            UInt64(): pl.UInt64,
            UInt32(): pl.UInt32,
            UInt16(): pl.UInt16,
            UInt8(): pl.UInt8,
            Float(): pl.Float64,  # we default to 64 bit
            Float64(): pl.Float64,
            Float32(): pl.Float32,
            Decimal(): pl.Decimal(scale=10),  # Arbitrary precision
            String(): pl.Utf8,
            Bool(): pl.Boolean,
            Datetime(): pl.Datetime("us"),
            Duration(): pl.Duration("us"),
            Time(): pl.Time,  # Polars uses nanoseconds since midnight
            Date(): pl.Date,
            NullType(): pl.Null,
        }[self]


class Float(Dtype):
    @classmethod
    def is_float(cls):
        return True


class Float64(Float): ...


class Float32(Float): ...


class Decimal(Float): ...


class Int(Dtype):
    @classmethod
    def is_int(cls):
        return True


class Int64(Int): ...


class Int32(Int): ...


class Int16(Int): ...


class Int8(Int): ...


class UInt64(Int): ...


class UInt32(Int): ...


class UInt16(Int): ...


class UInt8(Int): ...


class String(Dtype): ...


class Bool(Dtype): ...


class Datetime(Dtype): ...


class Date(Dtype): ...


class Time(Dtype): ...


class Duration(Dtype): ...


class NullType(Dtype): ...


class List(Dtype):
    def __init__(self, inner: "Dtype"):
        self.inner = inner

    def __eq__(self, rhs):
        return isinstance(rhs, List) and self.inner == rhs.inner

    def __hash__(self):
        return hash((0, hash(self.inner)))

    def __repr__(self):
        return f"List[{repr(self.inner)}]"

    def to_sql(self):
        import sqlalchemy as sqa

        return sqa.ARRAY(self.inner.to_sql())

    def to_polars(self):
        import polars as pl

        return pl.List(self.inner.to_polars())

    def to_arrow(self):
        import pyarrow as pa

        return pa.list_(self.inner.to_arrow())


class Enum(String):
    def __init__(self, *categories: str):
        if not all(isinstance(c, str) for c in categories):
            raise TypeError("arguments for `Enum` must have type `str`")
        self.categories = list(categories)

    def __eq__(self, rhs):
        return isinstance(rhs, Enum) and self.categories == rhs.categories

    def __repr__(self) -> str:
        return f"Enum[{', '.join(repr(c) for c in self.categories)}]"

    def __hash__(self):
        return hash(tuple(self.categories))

    def to_polars(self):
        import polars as pl

        return pl.Enum(self.categories)

    def to_sql(self):
        import sqlalchemy as sqa

        return sqa.String()

    def to_arrow(self):
        import pyarrow as pa

        # enum categories can only be maintained in pyarrow field (see to_arrow_field)
        return pa.string()

    def to_arrow_field(self, name: str, nullable: bool = True):
        """Convert this Dtype to a PyArrow Field."""
        import pyarrow as pa

        # try to mimic what polars does
        return pa.field(
            name,
            pa.dictionary(pa.uint32(), pa.large_string()),
            nullable=nullable,
            metadata={
                "_PL_ENUM_VALUES": "".join([f"{len(c)};{c}" for c in self.categories])
            },
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect

import pydiverse.common.dtypes as dtypes

ALL_TYPES = [
    getattr(dtypes, c)
    for c in dir(dtypes)
    if inspect.isclass(getattr(dtypes, c))
    and issubclass(getattr(dtypes, c), dtypes.Dtype)
    and c != "Dtype"
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import pytest

import pydiverse.common as pdc
from pydiverse.common.testing import ALL_TYPES

try:
    import pyarrow as pa
except ImportError:
    pa = None

from pydiverse.common import (
    Bool,
    Date,
    Datetime,
    Dtype,
    Enum,
    Float,
    Float32,
    Float64,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    String,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)


@pytest.mark.skipif(pa is None, reason="requires pyarrow")
def test_dtype_from_pyarrow():
    def assert_conversion(type_, expected):
        assert Dtype.from_arrow(type_) == expected

    assert_conversion(pa.int64(), Int64())
    assert_conversion(pa.int32(), Int32())
    assert_conversion(pa.int16(), Int16())
    assert_conversion(pa.int8(), Int8())

    assert_conversion(pa.uint64(), UInt64())
    assert_conversion(pa.uint32(), UInt32())
    assert_conversion(pa.uint16(), UInt16())
    assert_conversion(pa.uint8(), UInt8())

    assert_conversion(pa.float64(), Float64())
    assert_conversion(pa.float32(), Float32())
    assert_conversion(pa.float16(), Float32())

    assert_conversion(pa.string(), String())
    assert_conversion(pa.bool_(), Bool())

    assert_conversion(pa.date32(), Date())
    assert_conversion(pa.date64(), Date())

    assert_conversion(pa.time32("s"), Time())
    assert_conversion(pa.time32("ms"), Time())
    assert_conversion(pa.time64("us"), Time())
    assert_conversion(pa.time64("ns"), Time())

    assert_conversion(pa.timestamp("s"), Datetime())
    assert_conversion(pa.timestamp("ms"), Datetime())
    assert_conversion(pa.timestamp("us"), Datetime())
    assert_conversion(pa.timestamp("ns"), Datetime())


@pytest.mark.skipif(pa is None, reason="requires pyarrow")
def test_dtype_to_pyarrow():
    def assert_conversion(type_: Dtype, expected):
        assert type_.to_arrow() == expected

    assert_conversion(Int64(), pa.int64())
    assert_conversion(Int32(), pa.int32())
    assert_conversion(Int16(), pa.int16())
    assert_conversion(Int8(), pa.int8())

    assert_conversion(UInt64(), pa.uint64())
    assert_conversion(UInt32(), pa.uint32())
    assert_conversion(UInt16(), pa.uint16())
    assert_conversion(UInt8(), pa.uint8())

    assert_conversion(Float64(), pa.float64())
    assert_conversion(Float32(), pa.float32())

    assert_conversion(String(), pa.string())
    assert_conversion(Bool(), pa.bool_())

    assert_conversion(Date(), pa.date32())
    assert_conversion(Time(), pa.time64("us"))
    assert_conversion(Datetime(), pa.timestamp("us"))


@pytest.mark.skipif(pa is None, reason="requires pyarrow")
def test_dtype_to_pyarrow_enum():
    import polars as pl

    def assert_conversion(type_: Dtype, expected_dtype):
        df = pl.DataFrame(dict(x=["a"]), schema=dict(x=expected_dtype))
        expected = df.to_arrow().schema
        actual = pa.schema([type_.to_arrow_field("x", nullable=True)])
        assert actual == expected
        assert actual.field(0).metadata == expected.field(0).metadata

    assert_conversion(Enum("a", "b;c"), pl.Enum(["a", "b;c"]))


@pytest.mark.skipif(pa is None, reason="requires pyarrow")
@pytest.mark.parametrize(
    "type_",
    ALL_TYPES,
)
def test_all_types(type_):
    if type_ is pdc.List:
        type_obj = type_(pdc.Int64())
    elif type_ is pdc.Enum:
        type_obj = type_("a", "b", "c")
    else:
        type_obj = type_()

    acceptance_map = {
        Enum: String(),
        Float: Float64(),
        Int: Int64(),
    }

    dst_type = type_obj.to_arrow()
    back_type = Dtype.from_arrow(dst_type)

    assert back_type == acceptance_map.get(type_, type_obj)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import base64
import hashlib
import types
import warnings

try:
    import polars as pl
except ImportError:
    pl = types.ModuleType("pl")
    pl.DataFrame = None


def stable_hash(*args: str) -> str:
    """Compute a hash over a set of strings

    :param args: Some strings from which to compute the cache key
    :return: A sha256 base32 digest, trimmed to 20 char length
    """

    combined_hash = hashlib.sha256(b"PYDIVERSE")
    for arg in args:
        arg_bytes = str(arg).encode("utf8")
        arg_bytes_len = len(arg_bytes).to_bytes(length=8, byteorder="big")

        combined_hash.update(arg_bytes_len)
        combined_hash.update(arg_bytes)

    # Only take first 20 characters of base32 digest (100 bits). This
    # provides 50 bits of collision resistance, which is more than enough.
    # To illustrate: If you were to generate 1k hashes per second,
    # you still would have to wait over 800k years until you encounter
    # a collision.

    # NOTE: Can't use base64 because it contains lower and upper case
    #       letters; identifiers in pipedag are all lowercase
    hash_digest = combined_hash.digest()
    hash_str = base64.b32encode(hash_digest).decode("ascii").lower()
    return hash_str[:20]


def hash_polars_dataframe(df: pl.DataFrame, use_init_repr=False) -> str:
    if not use_init_repr:
        try:

            def unnest_all(df: pl.DataFrame) -> pl.DataFrame:
                while struct_cols_and_dtypes := [
                    (col, dtype)
                    for col, dtype in df.schema.items()
                    if dtype == pl.Struct
                ]:
                    df = df.with_columns(
                        pl.col(struct_col_name).struct.rename_fields(
                            [stable_hash(struct_col_name, struct_field_name)]
                        )
                        for struct_col_name, struct_field_name in struct_cols_and_dtypes
                    ).unnest(
                        struct_col_name for struct_col_name, _ in struct_cols_and_dtypes
                    )
                return df

            schema_hash = stable_hash(repr(df.schema))
            if df.is_empty():
                content_hash = "empty"
            else:
                # Since we need to operate on all lists, we need to access them first
                # if they are within a struct.
                df = unnest_all(df)
                array_columns = [
                    col for col, dtype in df.schema.items() if dtype == pl.Array
                ]
                list_columns = [
                    col for col, dtype in df.schema.items() if dtype == pl.List
                ]
                lf = df.lazy()
                if array_columns:
                    lf = lf.with_columns(pl.col(array_columns).reshape([-1]).implode())
                lf = lf.with_columns(
                    # Necessary because hash() does not work on lists of strings.
                    # This can be removed when
                    # https://github.com/pola-rs/polars/issues/21523 is resolved
                    # in all supported versions of polars.
                    pl.selectors.by_dtype(pl.List(pl.String)).list.eval(
                        pl.element().hash()
                    )
                )
                if list_columns or array_columns:
                    # Necessary because hash_rows() does not work on lists.
                    # This can be removed when
                    # https://github.com/pola-rs/polars/issues/24121 is resolved
                    # in all supported versions of polars.
                    lf = lf.with_columns(pl.col(*list_columns, *array_columns).hash())

                content_hash = str(
                    lf.collect()
                    .hash_rows()  # We get a Series of hashes, one for each row
                    # Since polars only hashes rows, we need to implode the Series into
                    # a single row to get a single hash
                    .implode()
                    .hash()
                    .item()
                )
            return "0" + stable_hash(schema_hash, content_hash)
        except Exception:
            warnings.warn(
                "Failed to compute hash for polars DataFrame in fast way. "
                "Falling back to to_init_repr() method.",
                stacklevel=1,
            )

    # fallback to to_init_repr string representation
    return "1" + stable_hash(df.to_init_repr(len(df)))



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import traceback
import types
from dataclasses import dataclass

import pytest

from pydiverse.common.errors import DisposedError
from pydiverse.common.util import Disposable, deep_map, requires
from pydiverse.common.util.hashing import hash_polars_dataframe

try:
    import polars as pl
except ImportError:
    pl = types.ModuleType("polars")
    pl.DataFrame = None


def test_requires():
    @requires(None, ImportError("Some Error"))
    class BadClass:
        a = 1
        b = 2

    # Shouldn't be able to create instance
    with pytest.raises(ImportError, match="Some Error"):
        BadClass()

    # Shouldn't be able to access class attribute
    with pytest.raises(ImportError, match="Some Error"):
        _ = BadClass.a

    # If all requirements are fulfilled, nothing should change
    @requires((pytest,), Exception("This shouldn't happen"))
    class GoodClass:
        a = 1

    _ = GoodClass()
    _ = GoodClass.a


def test_disposable():
    class Foo(Disposable):
        a = 1

        def bar(self):
            return 2

    x = Foo()

    assert x.a == 1
    assert x.bar() == 2

    x.dispose()

    with pytest.raises(DisposedError):
        _ = x.a
    with pytest.raises(DisposedError):
        x.foo()
    with pytest.raises(DisposedError):
        x.dispose()
    with pytest.raises(DisposedError):
        x.a = 1


def test_format_exception():
    # traceback.format_exception syntax changed from python 3.9 to 3.10
    # thus we use traceback.format_exc()
    try:
        raise RuntimeError("this error is intended by test")
    except RuntimeError:
        trace = traceback.format_exc()
        assert 'RuntimeError("this error is intended by test")' in trace
        assert "test_util.py" in trace


@dataclass
class Foo:
    a: int
    b: str
    c: list[int]
    d: tuple[int, str]


def test_deep_map():
    assert deep_map(1, lambda n: 2) == 2
    assert deep_map([1], lambda n: 0) == 0
    assert deep_map([1], lambda x: x) == [1]
    assert deep_map([3, 1, None, [3, 1, None, 4], 4], lambda x: 2 if x == 1 else x) == [
        3,
        2,
        None,
        [3, 2, None, 4],
        4,
    ]
    for outer in list, tuple:
        for inner in list, tuple:
            assert deep_map(
                outer([None, 3, 1, inner([3, 1, 4, None]), 4]),
                lambda x: 2 if x == 1 else x,
            ) == outer([None, 3, 2, inner([3, 2, 4, None]), 4])
    # attention: the replacement morphs keys 1 and 2;
    # the latter value overrides but is itself changed from 1 to 2 => 2:2
    assert deep_map(
        {1: 3, 2: 1, 3: None, 4: [3, 1, None, 4], 5: 4}, lambda x: 2 if x == 1 else x
    ) == {2: 2, 3: None, 4: [3, 2, None, 4], 5: 4}
    assert deep_map(
        dict(a=3, b=1, c=None, d=[3, 1, None, 4], e=4), lambda x: 2 if x == 1 else x
    ) == dict(a=3, b=2, c=None, d=[3, 2, None, 4], e=4)
    assert deep_map(
        Foo(1, "test", [1, 3], (1, "four")), lambda x: 2 if x == 1 else x
    ) == Foo(2, "test", [2, 3], (2, "four"))
    assert deep_map(
        [Foo(1, "test", [1, 3], (1, "four"))], lambda x: 2 if x == 1 else x
    ) == [Foo(2, "test", [2, 3], (2, "four"))]

    # Currently, deep_map cannot traverse other Iterables than lists, tuples, and dicts.
    d = {1: 1}
    res = deep_map([1, d.values()], lambda x: 2 if x == 1 else x)
    assert res[0] == 2
    assert list(res[1]) == list(d.values())


def check_df_hashes(df1_a: pl.DataFrame, other_dfs: list[pl.DataFrame]) -> None:
    assert hash_polars_dataframe(df1_a)[0] == "0"
    assert hash_polars_dataframe(df1_a, use_init_repr=True)[0] == "1"
    assert hash_polars_dataframe(df1_a) == hash_polars_dataframe(df1_a)
    assert hash_polars_dataframe(df1_a, use_init_repr=True) == hash_polars_dataframe(
        df1_a, use_init_repr=True
    )
    for df1_other in other_dfs:
        assert hash_polars_dataframe(df1_other)[0] == "0"
        assert hash_polars_dataframe(df1_other, use_init_repr=True)[0] == "1"
        assert hash_polars_dataframe(df1_a) != hash_polars_dataframe(df1_other)
        assert hash_polars_dataframe(
            df1_a, use_init_repr=True
        ) != hash_polars_dataframe(df1_other, use_init_repr=True)
        assert hash_polars_dataframe(df1_other) == hash_polars_dataframe(df1_other)
        assert hash_polars_dataframe(
            df1_other, use_init_repr=True
        ) == hash_polars_dataframe(df1_other, use_init_repr=True)


@pytest.mark.skipif(pl.DataFrame is None, reason="requires polars")
def test_hashing_basic():
    df1_a = pl.DataFrame(dict(x=[1]))
    df1_b = pl.DataFrame(dict(y=[1]))
    df1_c = pl.DataFrame(dict(x=[2]))
    df1_d = pl.DataFrame(dict(x=[1.0]))
    df1_e = pl.DataFrame(dict(x=[]))

    check_df_hashes(df1_a, [df1_b, df1_c, df1_d, df1_e])


@pytest.mark.skipif(pl.DataFrame is None, reason="requires polars")
def test_hashing():
    df1_a = pl.DataFrame(
        data=dict(x=[["foo", "bar"], [""]], y=[[1, 2], None], z=[1, 2])
    ).with_columns(s=pl.struct("x", "y"))
    df1_b = pl.DataFrame(
        data=dict(x=[["foo", "bar"], [""]], z=[[1, 2], None], y=[1, 2])
    ).with_columns(s=pl.struct("x", "y"))
    df1_c = pl.DataFrame(
        data=dict(x=[["foo", "baR"], [""]], y=[[1, 2], None], z=[1, 2])
    ).with_columns(s=pl.struct("x", "y"))
    df1_d = pl.DataFrame(
        data=dict(x=[["foo", "bar"], [""]], y=[[1, 3], None], z=[1, 2])
    ).with_columns(s=pl.struct("x", "y"))
    df1_e = pl.DataFrame(
        data=dict(x=[["foo", "bar"], [""]], y=[[1, 3], []], z=[1, 2])
    ).with_columns(s=pl.struct("x", "y"))
    df1_f = pl.DataFrame(
        data=dict(x=[["foo", "bar"], [""]], y=[[1, 2], None], z=[1, 2])
    ).with_columns(s=pl.struct("x", "z"))

    check_df_hashes(df1_a, [df1_b, df1_c, df1_d, df1_e, df1_f])


@pytest.mark.skipif(pl.DataFrame is None, reason="requires polars")
def test_hashing_array():
    df1_a = pl.DataFrame(
        data=dict(x=[[[1], [2], [3]]]), schema=dict(x=pl.Array(pl.UInt16, shape=(3, 1)))
    )
    df1_b = pl.DataFrame(
        data=dict(y=[[[1], [2], [3]]]), schema=dict(y=pl.Array(pl.UInt16, shape=(3, 1)))
    )
    df1_c = pl.DataFrame(
        data=dict(x=[[[1], [3], [2]]]), schema=dict(x=pl.Array(pl.UInt16, shape=(3, 1)))
    )
    df1_d = pl.DataFrame(
        data=dict(x=[[[1, 2, 3]]]), schema=dict(x=pl.Array(pl.UInt16, shape=(1, 3)))
    )
    df1_e = pl.DataFrame(
        data=dict(x=[[1, 2, 3]]), schema=dict(x=pl.Array(pl.UInt16, shape=3))
    )

    check_df_hashes(df1_a, [df1_b, df1_c, df1_d, df1_e])



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause


class DisposedError(Exception):
    """
    Exception raise when an object has been disposed, but some attributes are
    being accessed nevertheless.
    """



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
from .deep_map import deep_map
from .deep_merge import deep_merge
from .disposable import Disposable
from .import_ import requires

__all__ = [
    "deep_map",
    "deep_merge",
    "Disposable",
    "requires",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import datetime as dt
import types

import pytest

import pydiverse.common as pdc
from pydiverse.common.testing import ALL_TYPES

try:
    import numpy as np
    import pandas as pd
    import pyarrow as pa
except ImportError:
    np = pa = None
    pd = types.ModuleType("pandas")
    pd.__version__ = "0.0.0"

from pydiverse.common import (
    Bool,
    Date,
    Datetime,
    Decimal,
    Dtype,
    Enum,
    Float,
    Float32,
    Float64,
    Int,
    Int8,
    Int16,
    Int32,
    Int64,
    PandasBackend,
    String,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)


@pytest.mark.skipif(np is None, reason="requires pandas, numpy, and pyarrow")
def test_dtype_from_pandas():
    def assert_conversion(type_, expected):
        assert Dtype.from_pandas(type_) == expected

    assert_conversion(int, Int64())
    assert_conversion(float, Float64())
    assert_conversion(str, String())
    assert_conversion(bool, Bool())

    # Objects should get converted to string
    assert_conversion(object, String())
    assert_conversion(dt.date, String())
    assert_conversion(dt.time, String())
    assert_conversion(dt.datetime, String())

    # Numpy types
    assert_conversion(np.int64, Int64())
    assert_conversion(np.int32, Int32())
    assert_conversion(np.int16, Int16())
    assert_conversion(np.int8, Int8())

    assert_conversion(np.uint64, UInt64())
    assert_conversion(np.uint32, UInt32())
    assert_conversion(np.uint16, UInt16())
    assert_conversion(np.uint8, UInt8())

    assert_conversion(np.floating, Float64())
    assert_conversion(np.float64, Float64())
    assert_conversion(np.float32, Float32())

    assert_conversion(np.bytes_, String())
    assert_conversion(np.bool_, Bool())

    assert_conversion(np.datetime64, Datetime())
    assert_conversion(np.dtype("datetime64[ms]"), Datetime())
    assert_conversion(np.dtype("datetime64[ns]"), Datetime())

    # Numpy nullable extension types
    assert_conversion(pd.Int64Dtype(), Int64())
    assert_conversion(pd.Int32Dtype(), Int32())
    assert_conversion(pd.Int16Dtype(), Int16())
    assert_conversion(pd.Int8Dtype(), Int8())

    assert_conversion(pd.UInt64Dtype(), UInt64())
    assert_conversion(pd.UInt32Dtype(), UInt32())
    assert_conversion(pd.UInt16Dtype(), UInt16())
    assert_conversion(pd.UInt8Dtype(), UInt8())

    assert_conversion(pd.Float64Dtype(), Float64())
    assert_conversion(pd.Float32Dtype(), Float32())

    assert_conversion(pd.StringDtype(), String())
    assert_conversion(pd.BooleanDtype(), Bool())


@pytest.mark.skipif(np is None, reason="requires pandas, numpy, and pyarrow")
def test_dtype_to_pandas_numpy():
    def assert_conversion(type_: Dtype, expected):
        assert type_.to_pandas(PandasBackend.NUMPY) == expected

    assert_conversion(Int64(), pd.Int64Dtype())
    assert_conversion(Int32(), pd.Int32Dtype())
    assert_conversion(Int16(), pd.Int16Dtype())
    assert_conversion(Int8(), pd.Int8Dtype())

    assert_conversion(UInt64(), pd.UInt64Dtype())
    assert_conversion(UInt32(), pd.UInt32Dtype())
    assert_conversion(UInt16(), pd.UInt16Dtype())
    assert_conversion(UInt8(), pd.UInt8Dtype())

    assert_conversion(String(), pd.StringDtype())
    assert_conversion(Bool(), pd.BooleanDtype())

    assert_conversion(Date(), "datetime64[s]")
    assert_conversion(Datetime(), "datetime64[us]")

    with pytest.raises(TypeError):
        Time().to_pandas(PandasBackend.NUMPY)


@pytest.mark.skipif(np is None, reason="requires pandas, numpy, and pyarrow")
@pytest.mark.skipif("pd.__version__ < '2'")
def test_dtype_to_pandas_pyarrow():
    def assert_conversion(type_: Dtype, expected):
        if isinstance(expected, pa.DataType):
            assert type_.to_pandas(PandasBackend.ARROW) == pd.ArrowDtype(expected)
        else:
            assert type_.to_pandas(PandasBackend.ARROW) == expected

    assert_conversion(Int64(), pa.int64())
    assert_conversion(Int32(), pa.int32())
    assert_conversion(Int16(), pa.int16())
    assert_conversion(Int8(), pa.int8())

    assert_conversion(UInt64(), pa.uint64())
    assert_conversion(UInt32(), pa.uint32())
    assert_conversion(UInt16(), pa.uint16())
    assert_conversion(UInt8(), pa.uint8())

    assert_conversion(String(), pd.StringDtype(storage="pyarrow"))
    assert_conversion(Bool(), pa.bool_())

    assert_conversion(Date(), pa.date32())
    assert_conversion(Time(), pa.time64("us"))
    assert_conversion(Datetime(), pa.timestamp("us"))


@pytest.mark.skipif(np is None, reason="requires pandas, numpy, and pyarrow")
@pytest.mark.parametrize(
    "type_",
    ALL_TYPES,
)
def test_all_types_numpy(type_):
    if type_ is pdc.List:
        type_obj = type_(pdc.Int64())
    elif type_ is pdc.Enum:
        type_obj = type_("a", "b", "c")
    else:
        type_obj = type_()

    if type_ is pdc.NullType:
        with pytest.raises(TypeError, match="pandas doesn't have a native null dtype"):
            type_obj.to_pandas(PandasBackend.NUMPY)
    elif type_ is pdc.Time:
        with pytest.raises(TypeError, match="pandas doesn't have a native time dtype"):
            type_obj.to_pandas(PandasBackend.NUMPY)
    elif type_ is pdc.List:
        with pytest.raises(TypeError, match="pandas doesn't have a native list dtype"):
            type_obj.to_pandas(PandasBackend.NUMPY)
    else:
        acceptance_map = {
            Float: Float64(),
            Int: Int64(),
            Decimal: Float64(),
            Date: Datetime(),
        }

        dst_type = type_obj.to_pandas(PandasBackend.NUMPY)
        back_type = Dtype.from_pandas(dst_type)

        assert back_type == acceptance_map.get(type_, type_obj)


@pytest.mark.skipif(np is None, reason="requires pandas, numpy, and pyarrow")
@pytest.mark.parametrize(
    "type_",
    ALL_TYPES,
)
def test_all_types_arrow(type_):
    if type_ is pdc.List:
        type_obj = type_(pdc.Int64())
    elif type_ is pdc.Enum:
        type_obj = type_("a", "b", "c")
    else:
        type_obj = type_()

    acceptance_map = {
        Enum: String(),
        Float: Float64(),
        Int: Int64(),
    }

    dst_type = type_obj.to_pandas(PandasBackend.ARROW)
    back_type = Dtype.from_pandas(dst_type)
    assert back_type == acceptance_map.get(type_, type_obj)
